@@ -1,21 +1,80 @@
 import abc
+import gzip
+import io
 import logging
-import os
-import re
 import tempfile
 import time
-from urllib.error import HTTPError
-from urllib.request import urlopen
+import uuid
 
+import msgpack
+import numpy as np
 import pandas as pd
+from tdclient.util import normalized_msgpack
+
+from .spark import fetch_td_spark_context
 
-TD_SPARK_BASE_URL = "https://s3.amazonaws.com/td-spark/{}"
-TD_SPARK_JAR_NAME = "td-spark-assembly_2.11-19.7.0.jar"
 logger = logging.getLogger(__name__)
 
+if pd.__version__ >= "1.0.0":
+    logger.warning(
+        "You are using pandas 1.0.0 or later, but pytd is not fully tested on "
+        "the major version. Since the version has introduced some "
+        "experimental, backward-incompatible features, there might be a "
+        "chance of unexpected runtime failure when you handle a DataFrame "
+        "containing null or array. See pandas official documentation for more "
+        "information: "
+        "https://pandas.pydata.org/pandas-docs/version/1.0.0/whatsnew/v1.0.0.html"
+    )
 
-def _cast_dtypes(dataframe, inplace=True):
-    """Convert dtypes into one of {int, float, str} type.
+
+def _is_np_nan(x):
+    return isinstance(x, float) and np.isnan(x)
+
+
+def _is_0d_ary(x):
+    return isinstance(x, np.ndarray) and len(x.shape) == 0
+
+
+def _is_0d_nan(x):
+    return _is_0d_ary(x) and x.dtype.kind == "f" and np.isnan(x)
+
+
+def _isnull(x):
+    return x is None or _is_np_nan(x)
+
+
+def _isinstance_or_null(x, t):
+    return _isnull(x) or isinstance(x, t)
+
+
+def _to_list(ary):
+    # Return None if None, np.nan, or np.nan in 0-d array given
+    if ary is None or _is_np_nan(ary) or _is_0d_nan(ary):
+        return None
+
+    # Return Python primitive value if 0-d array given
+    if _is_0d_ary(ary):
+        return ary.tolist()
+
+    _ary = np.asarray(ary)
+    # Replace numpy.nan to None which will be converted to NULL on TD
+    kind = _ary.dtype.kind
+    if kind == "f":
+        _ary = np.where(np.isnan(_ary), None, _ary)
+    elif kind == "U":
+        _ary = np.where(_ary == "nan", None, _ary)
+    elif kind == "O":
+        _ary = np.array([None if _is_np_nan(x) else x for x in _ary])
+    return _ary.tolist()
+
+
+def _convert_nullable_str(x, t, lower=False):
+    v = str(x).lower() if lower else str(x)
+    return v if isinstance(x, t) else None
+
+
+def _cast_dtypes(dataframe, inplace=True, keep_list=False):
+    """Convert dtypes into one of {int, float, str, list} type.
 
     A character code (one of ‘biufcmMOSUV’) identifying the general kind of data.
 
@@ -34,12 +93,28 @@ def _cast_dtypes(dataframe, inplace=True):
     df = dataframe if inplace else dataframe.copy()
 
     for column, kind in dataframe.dtypes.apply(lambda dtype: dtype.kind).iteritems():
-        if kind == "i" or kind == "u":
+        t = str
+        if kind in ("i", "u"):
             t = "Int64" if df[column].isnull().any() else "int64"
         elif kind == "f":
             t = float
-        else:
-            t = str
+        elif kind == "O":
+            t = object
+            if df[column].apply(_isinstance_or_null, args=((list, np.ndarray),)).all():
+                if keep_list:
+                    df[column] = df[column].apply(_to_list)
+                else:
+                    df[column] = df[column].apply(
+                        _convert_nullable_str, args=((list, np.ndarray),)
+                    )
+            elif df[column].apply(_isinstance_or_null, args=(bool,)).all():
+                df[column] = df[column].apply(
+                    _convert_nullable_str, args=(bool,), lower=True
+                )
+            elif df[column].apply(_isinstance_or_null, args=(str,)).all():
+                df[column] = df[column].apply(_convert_nullable_str, args=(str,))
+            else:
+                t = str
         df[column] = df[column].astype(t)
 
         # Bulk Import API internally handles boolean string as a boolean type,
@@ -108,14 +183,15 @@ class InsertIntoWriter(Writer):
 
         Parameters
         ----------
-        dataframe : pandas.DataFrame
+        dataframe : :class:`pandas.DataFrame`
             Data loaded to a target table.
 
-        table : pytd.table.Table
+        table : :class:`pytd.table.Table`
             Target table.
 
         if_exists : {'error', 'overwrite', 'append', 'ignore'}
             What happens when a target table already exists.
+
             - error: raise an exception.
             - overwrite: drop it, recreate it, and insert data.
             - append: insert data. Create if does not exist.
@@ -141,30 +217,31 @@ class InsertIntoWriter(Writer):
 
         Parameters
         ----------
-        table : pytd.table.Table
+        table : :class:`pytd.table.Table`
             Target table.
 
         list_of_tuple : list of tuples
             Data loaded to a target table. Each element is a tuple that
             represents single table row.
 
-        column_names : list of string
+        column_names : list of str
             Column names.
 
-        column_types : list of string
+        column_types : list of str
             Column types corresponding to the names. Note that Treasure Data
             supports limited amount of types as documented in:
             https://support.treasuredata.com/hc/en-us/articles/360001266468-Schema-Management
 
         if_exists : {'error', 'overwrite', 'append', 'ignore'}
             What happens when a target table already exists.
+
             - error: raise an exception.
             - overwrite: drop it, recreate it, and insert data.
             - append: insert data. Create if does not exist.
             - ignore: do nothing.
         """
 
-        if table.exist:
+        if table.exists:
             if if_exists == "error":
                 raise RuntimeError(
                     "target table '{}.{}' already exists".format(
@@ -194,17 +271,17 @@ class InsertIntoWriter(Writer):
 
         Parameters
         ----------
-        database : string
+        database : str
             Target database name.
 
-        table : string
+        table : str
             Target table name.
 
         list_of_tuple : list of tuples
             Data loaded to a target table. Each element is a tuple that
             represents single table row.
 
-        column_names : list of string
+        column_names : list of str
             Column names.
         """
         rows = []
@@ -229,27 +306,94 @@ class BulkImportWriter(Writer):
     td-client-python's bulk importer.
     """
 
-    def write_dataframe(self, dataframe, table, if_exists):
+    def write_dataframe(self, dataframe, table, if_exists, fmt="csv", keep_list=False):
         """Write a given DataFrame to a Treasure Data table.
 
-        This method internally converts a given pandas.DataFrame into a
-        temporary CSV file, and upload the file to Treasure Data via bulk
+        This method internally converts a given :class:`pandas.DataFrame` into a
+        temporary CSV/msgpack file, and upload the file to Treasure Data via bulk
         import API.
 
         Parameters
         ----------
-        dataframe : pandas.DataFrame
+        dataframe : :class:`pandas.DataFrame`
             Data loaded to a target table.
 
-        table : pytd.table.Table
+        table : :class:`pytd.table.Table`
             Target table.
 
         if_exists : {'error', 'overwrite', 'append', 'ignore'}
             What happens when a target table already exists.
+
             - error: raise an exception.
             - overwrite: drop it, recreate it, and insert data.
             - append: insert data. Create if does not exist.
             - ignore: do nothing.
+
+        fmt : {'csv', 'msgpack'}, default: 'csv'
+            Format for bulk_import.
+
+            - csv
+                Convert dataframe to temporary CSV file. Stable option but slower
+                than msgpack option because pytd saves dataframe as temporary CSV file,
+                then td-client converts it to msgpack.
+            - msgpack
+                Convert to temporary msgpack.gz file. Fast option but might be
+                unstable because of skipping ``tdclient.API._prepare_file()``
+
+        keep_list : boolean, default: False
+            If this argument is True, keep list or numpy.ndarray column as list, which
+            will be converted array<T> on Treasure Data table.
+            Each type of element of list will be converted by
+            ``numpy.array(your_list).tolist()``.
+
+            If True, ``fmt`` argument will be overwritten with ``msgpack``.
+
+            Examples
+            ---------
+
+            A dataframe containing list will be treated array<T> in TD.
+
+            >>> import pytd
+            >>> import numpy as np
+            >>> import pandas as pd
+            >>> df = pd.DataFrame(
+            ...     {
+            ...         "a": [[1, 2, 3], [2, 3, 4]],
+            ...         "b": [[0, None, 2], [2, 3, 4]],
+            ...         "c": [np.array([1, np.nan, 3]), [2, 3, 4]]
+            ...     }
+            ... )
+            >>> client = pytd.Client()
+            >>> table = pytd.table.Table(client, "mydb", "test")
+            >>> writer = pytd.writer.BulkImportWriter()
+            >>> writer.write_dataframe(df, table, if_exists="overwrite", keep_list=True)
+
+            In this case, the type of columns will be:
+            ``{"a": array<int>, "b": array<string>, "c": array<string>}``
+
+            If you want to set the type after ingestion, you need to run
+            ``tdclient.Client.update_schema`` like:
+
+            >>> client.api_client.update_schema(
+            ...     "mydb",
+            ...     "test",
+            ...     [
+            ...         ["a", "array<long>", "a"],
+            ...         ["b", "array<int>", "b"],
+            ...         ["c", "array<int>", "c"],
+            ...     ],
+            ... )
+
+            Note that ``numpy.nan`` will be converted as a string value as ``"NaN"`` or
+            ``"nan"``, so pytd will convert ``numpy.nan`` to ``None`` only when the
+            dtype of a ndarray is `float`.
+            Also, numpy converts integer array including ``numpy.nan`` into float array
+            because ``numpy.nan`` is a Floating Point Special Value. See also:
+            https://docs.scipy.org/doc/numpy-1.13.0/user/misc.html#ieee-754-floating-point-special-values
+
+            Or, you can use :func:`Client.load_table_from_dataframe` function as well.
+
+            >>> client.load_table_from_dataframe(df, "bulk_import", keep_list=True)
         """
         if self.closed:
             raise RuntimeError("this writer is already closed and no longer available")
@@ -257,37 +401,54 @@ class BulkImportWriter(Writer):
         if "time" not in dataframe.columns:  # need time column for bulk import
             dataframe["time"] = int(time.time())
 
-        fp = tempfile.NamedTemporaryFile(suffix=".csv")
+        # We enforce using "msgpack" format for list since CSV can't handle list.
+        if keep_list:
+            fmt = "msgpack"
 
-        _cast_dtypes(dataframe)
-        dataframe.to_csv(fp.name)
+        _cast_dtypes(dataframe, keep_list=keep_list)
 
-        self._bulk_import(table, fp, if_exists)
+        if fmt == "csv":
+            fp = tempfile.NamedTemporaryFile(suffix=".csv")
+            dataframe.to_csv(fp.name)
+        elif fmt == "msgpack":
+            fp = io.BytesIO()
+            fp = self._write_msgpack_stream(dataframe.to_dict(orient="records"), fp)
+        else:
+            raise ValueError(
+                "unsupported format '{}' for bulk import. "
+                "should be 'csv' or 'msgpack'".format(fmt)
+            )
+
+        self._bulk_import(table, fp, if_exists, fmt)
 
         fp.close()
 
-    def _bulk_import(self, table, csv, if_exists):
+    def _bulk_import(self, table, file_like, if_exists, fmt="csv"):
         """Write a specified CSV file to a Treasure Data table.
 
         This method uploads the file to Treasure Data via bulk import API.
 
         Parameters
         ----------
-        table : pytd.table.Table
+        table : :class:`pytd.table.Table`
             Target table.
 
-        csv : File pointer of a CSV file
+        file_like : File like object
             Data in this file will be loaded to a target table.
 
-        if_exists : {'error', 'overwrite', 'append', 'ignore'}
+        if_exists : str, {'error', 'overwrite', 'append', 'ignore'}
             What happens when a target table already exists.
+
             - error: raise an exception.
             - overwrite: drop it, recreate it, and insert data.
             - append: insert data. Create if does not exist.
             - ignore: do nothing.
+
+        fmt : str, optional, {'csv', 'msgpack'}, default: 'csv'
+            File format for bulk import. See also :func:`write_dataframe`
         """
         params = None
-        if table.exist:
+        if table.exists:
             if if_exists == "error":
                 raise RuntimeError(
                     "target table '{}.{}' already exists".format(
@@ -302,18 +463,23 @@ class BulkImportWriter(Writer):
                 table.delete()
                 table.create()
             else:
-                raise ValueError("invalid valud for if_exists: {}".format(if_exists))
+                raise ValueError("invalid value for if_exists: {}".format(if_exists))
         else:
             table.create()
 
-        session_name = "session-{}".format(int(time.time()))
+        session_name = "session-{}".format(uuid.uuid1())
 
         bulk_import = table.client.api_client.create_bulk_import(
             session_name, table.database, table.table, params=params
         )
         try:
-            logger.info("uploading data converted into a CSV file")
-            bulk_import.upload_file("part", "csv", csv.name)
+            logger.info("uploading data converted into a {} file".format(fmt))
+            if fmt == "msgpack":
+                size = file_like.getbuffer().nbytes
+                # To skip API._prepare_file(), which recreate msgpack again.
+                bulk_import.upload_part("part", file_like, size)
+            else:
+                bulk_import.upload_file("part", fmt, file_like)
             bulk_import.freeze()
         except Exception as e:
             bulk_import.delete()
@@ -344,21 +510,62 @@ class BulkImportWriter(Writer):
         bulk_import.commit(wait=True)
         bulk_import.delete()
 
+    def _write_msgpack_stream(self, items, stream):
+        """Write MessagePack stream
+
+        Parameters
+        ----------
+        items : list of dict
+            Same format with dataframe.to_dict(orient="records")
+            Examples:
+                ``[{"time": 12345, "col1": "foo"}, {"time": 12345, "col1": "bar"}]``
+        stream : File like object
+            Target file like object which has `write()` function. This object will be
+            updated in this function.
+        """
+
+        with gzip.GzipFile(mode="wb", fileobj=stream) as gz:
+            packer = msgpack.Packer()
+            for item in items:
+                try:
+                    mp = packer.pack(item)
+                except (OverflowError, ValueError):
+                    packer.reset()
+                    mp = packer.pack(normalized_msgpack(item))
+                gz.write(mp)
+
+        stream.seek(0)
+        return stream
+
 
 class SparkWriter(Writer):
     """A writer module that loads Python data to Treasure Data.
 
     Parameters
     ----------
-    td_spark_path : string, optional
+    td_spark_path : str, optional
         Path to td-spark-assembly_x.xx-x.x.x.jar. If not given, seek a path
-        ``__file__ + TD_SPARK_JAR_NAME`` by default.
+        ``TDSparkContextBuilder.default_jar_path()`` by default.
 
-    download_if_missing : boolean, default: True
+    download_if_missing : bool, default: True
         Download td-spark if it does not exist at the time of initialization.
 
     spark_configs : dict, optional
         Additional Spark configurations to be set via ``SparkConf``'s ``set`` method.
+
+    Attributes
+    ----------
+    td_spark_path : str
+        Path to td-spark-assembly_x.xx-x.x.x.jar.
+
+    download_if_missing : bool
+        Download td-spark if it does not exist at the time of initialization.
+
+    spark_configs : dict
+        Additional Spark configurations to be set via ``SparkConf``'s ``set`` method.
+
+    td_spark : :class:`td_pyspark.TDSparkContext`
+        Connection of td-spark
     """
 
     def __init__(
@@ -373,25 +580,26 @@ class SparkWriter(Writer):
 
     @property
     def closed(self):
-        return self.td_spark is not None and self.td_spark._jsc.sc().isStopped()
+        return self.td_spark is not None and self.td_spark.spark._jsc.sc().isStopped()
 
     def write_dataframe(self, dataframe, table, if_exists):
         """Write a given DataFrame to a Treasure Data table.
 
-        This method internally converts a given pandas.DataFrame into Spark
+        This method internally converts a given :class:`pandas.DataFrame` into Spark
         DataFrame, and directly writes to Treasure Data's main storage
         so-called Plazma through a PySpark session.
 
         Parameters
         ----------
-        dataframe : pandas.DataFrame
+        dataframe : :class:`pandas.DataFrame`
             Data loaded to a target table.
 
-        table : pytd.table.Table
+        table : :class:`pytd.table.Table`
             Target table.
 
         if_exists : {'error', 'overwrite', 'append', 'ignore'}
             What happens when a target table already exists.
+
             - error: raise an exception.
             - overwrite: drop it, recreate it, and insert data.
             - append: insert data. Create if does not exist.
@@ -401,16 +609,17 @@ class SparkWriter(Writer):
             raise RuntimeError("this writer is already closed and no longer available")
 
         if if_exists not in ("error", "overwrite", "append", "ignore"):
-            raise ValueError("invalid valud for if_exists: {}".format(if_exists))
+            raise ValueError("invalid value for if_exists: {}".format(if_exists))
 
         if self.td_spark is None:
-            self.td_spark = self._fetch_td_spark(
+            self.td_spark = fetch_td_spark_context(
                 table.client.apikey,
                 table.client.endpoint,
                 self.td_spark_path,
                 self.download_if_missing,
                 self.spark_configs,
             )
+
             self.fetched_apikey, self.fetched_endpoint = (
                 table.client.apikey,
                 table.client.endpoint,
@@ -427,11 +636,13 @@ class SparkWriter(Writer):
         from py4j.protocol import Py4JJavaError
 
         _cast_dtypes(dataframe)
-        sdf = self.td_spark.createDataFrame(dataframe)
+        # Replace np.nan to None to avoid Int64 conversion issue
+        if dataframe.isnull().any().any():
+            dataframe.replace({np.nan: None}, inplace=True)
+        sdf = self.td_spark.spark.createDataFrame(dataframe)
         try:
-            sdf.write.mode(if_exists).format("com.treasuredata.spark").option(
-                "table", "{}.{}".format(table.database, table.table)
-            ).save()
+            destination = "{}.{}".format(table.database, table.table)
+            self.td_spark.write(sdf, destination, if_exists)
         except Py4JJavaError as e:
             if "API_ACCESS_FAILURE" in str(e.java_exception):
                 raise PermissionError(
@@ -446,78 +657,4 @@ class SparkWriter(Writer):
         """Close a PySpark session connected to Treasure Data.
         """
         if self.td_spark is not None:
-            self.td_spark.stop()
-
-    def _fetch_td_spark(
-        self, apikey, endpoint, td_spark_path, download_if_missing, spark_configs
-    ):
-        try:
-            from pyspark.conf import SparkConf
-            from pyspark.sql import SparkSession
-        except ImportError:
-            raise RuntimeError("PySpark is not installed")
-
-        conf = (
-            SparkConf()
-            .setMaster("local[*]")
-            .set("spark.serializer", "org.apache.spark.serializer.KryoSerializer")
-            .set("spark.sql.execution.arrow.enabled", "true")
-        )
-        conf.set("spark.td.apikey", apikey)
-
-        if td_spark_path is None:
-            td_spark_path = os.path.join(
-                os.path.dirname(os.path.abspath(__file__)), TD_SPARK_JAR_NAME
-            )
-
-        available = os.path.exists(td_spark_path)
-
-        if not available and download_if_missing:
-            self._download_td_spark(td_spark_path)
-        elif not available:
-            raise IOError("td-spark is not found and `download_if_missing` is False")
-
-        conf.set("spark.jars", td_spark_path)
-
-        plazma_api = os.getenv("TD_PLAZMA_API")
-        presto_api = os.getenv("TD_PRESTO_API")
-
-        if plazma_api and presto_api:
-            api_regex = re.compile(r"(?:https?://)?(api(?:-.+?)?)\.")
-            conf.set("spark.td.api.host", api_regex.sub("\\1.", endpoint).strip("/"))
-            conf.set("spark.td.plazma_api.host", plazma_api)
-            conf.set("spark.td.presto_api.host", presto_api)
-
-        site = "us"
-        if ".co.jp" in endpoint:
-            site = "jp"
-        if "eu01" in endpoint:
-            site = "eu01"
-        conf.set("spark.td.site", site)
-
-        if isinstance(spark_configs, dict):
-            for k, v in spark_configs.items():
-                conf.set(k, v)
-
-        try:
-            return SparkSession.builder.config(conf=conf).getOrCreate()
-        except Exception as e:
-            raise RuntimeError("failed to connect to td-spark: " + str(e))
-
-    def _download_td_spark(self, destination):
-        download_url = TD_SPARK_BASE_URL.format(TD_SPARK_JAR_NAME)
-        try:
-            response = urlopen(download_url)
-        except HTTPError:
-            raise RuntimeError("failed to access to the download URL: " + download_url)
-
-        logger.info("Downloading td-spark...")
-        try:
-            with open(destination, "w+b") as f:
-                f.write(response.read())
-        except Exception:
-            os.remove(destination)
-            raise
-        logger.info("Completed to download")
-
-        response.close()
+            self.td_spark.spark.stop()
